__all__ = ["Polyline", "Plane", "Box", "Line", "CompositeTransform",  "CoordinateManager"]

from .package_version import __version__

from .box._box_object import Box
from .plane._plane_object import Plane
from .line._line_object import Line
from .polyline._polyline_object import Polyline
from .transform._composite_transform import CompositeTransform
from .transform._coordinate_manager import CoordinateManager



import numpy as np
import vg


def subdivide_segment(p1, p2, num_points, endpoint=True):
    """
    For two points in n-space, return an np.ndarray of equidistant partition
    points along the segment determined by p1 & p2.

    The total number of points returned will be n_samples. When n_samples is
    2, returns the original points.

    When endpoint is True, p2 is the last point. When false, p2 is excluded.

    Partition order is oriented from p1 to p2.

    Args:
        p1, p2:
            1 x N vectors

        partition_size:
            size of partition. should be >= 2.

    """
    if not isinstance(num_points, int):
        raise TypeError("partition_size should be an int.")
    elif num_points < 2:
        raise ValueError("partition_size should be bigger than 1.")

    return (p2 - p1) * np.linspace(0, 1, num=num_points, endpoint=endpoint)[
        :, np.newaxis
    ] + p1


def subdivide_segments(v, num_subdivisions=5):
    """

    params:
        v:
            V x N np.array of points in N-space

        partition_size:
            how many partitions intervals for each segment?

    Fill in the line segments determined by v with equally
    spaced points - the space for each segment is determined
    by the length of the segment and the supplied partition size.

    """
    src = np.arange(len(v) - 1)
    dst = src + 1

    diffs = v[dst] - v[src]

    sqdis = np.square(diffs)
    dists = np.sqrt(np.sum(sqdis, axis=1))

    unitds = diffs / dists[:, np.newaxis]
    widths = dists / num_subdivisions

    domain = widths[:, np.newaxis] * np.arange(0, num_subdivisions)
    domain = domain.flatten()[:, np.newaxis]

    points = np.repeat(v[:-1], num_subdivisions, axis=0)
    unitds = np.repeat(unitds, num_subdivisions, axis=0)

    filled = points + (unitds * domain)

    return np.vstack((filled, v[-1]))


def closest_point_of_line_segment(points, start_points, segment_vectors):
    """
    Compute pairwise the point on each line segment that is nearest to the
    corresponding query point.
    """
    # Adapted from public domain algorithm
    # https://gdbooks.gitbooks.io/3dcollisions/content/Chapter1/closest_point_on_line.html
    k = vg.shape.check(locals(), "points", (-1, 3))
    vg.shape.check(locals(), "start_points", (k, 3))
    vg.shape.check(locals(), "segment_vectors", (k, 3))

    # Compute t values such that
    # `result = reference_point + t * vector_along_line`.
    square_of_segment_lengths = vg.dot(segment_vectors, segment_vectors)
    # Degenerate segments will cause a division by zero, so handle that.
    t = np.nan_to_num(
        vg.dot(points - start_points, segment_vectors) / square_of_segment_lengths
    )

    # When `0 <= t <= 1`, the point is on the segment. When `t < 0`, the
    # closest point is the segment start. When `t > 1`, the closest point is
    # the segment end.
    #
    # Start with the `0 <= t <= 1 case`, then use masks to apply the clamp.
    result = start_points + t.reshape(-1, 1) * segment_vectors

    clamped_to_start_point = t < 0
    result[clamped_to_start_point] = start_points[clamped_to_start_point]

    clamped_to_end_point = t > 1
    result[clamped_to_end_point] = (
        start_points[clamped_to_end_point] + segment_vectors[clamped_to_end_point]
    )

    return result



import vg
from .._common.shape import check_shape_any

__all__ = ["project_point_to_line", "coplanar_points_are_on_same_side_of_line"]


def project_point_to_line(points, reference_points_of_lines, vectors_along_lines):
    """
    Project a point to a line, or pairwise project a stack of points to a
    stack of lines.
    """
    k = check_shape_any(points, (3,), (-1, 3), name="points")
    check_shape_any(
        reference_points_of_lines,
        (3,),
        (-1 if k is None else k, 3),
        name="reference_points_of_lines",
    )
    vg.shape.check(locals(), "vectors_along_lines", reference_points_of_lines.shape)

    return reference_points_of_lines + vg.project(
        points - reference_points_of_lines, onto=vectors_along_lines
    )


def coplanar_points_are_on_same_side_of_line(a, b, p1, p2):
    """
    Test if the given points are on the same side of the given line.

    Args:
        a (np.arraylike): The first 3D point of interest.
        b (np.arraylike): The second 3D point of interest.
        p1 (np.arraylike): A first point which lies on the line of interest.
        p2 (np.arraylike): A second point which lies on the line of interest.

    Returns:
        bool: `True` when `a` and `b` are on the same side of the line defined
        by `p1` and `p2`.
    """
    vg.shape.check(locals(), "a", (3,))
    vg.shape.check(locals(), "b", (3,))
    vg.shape.check(locals(), "p1", (3,))
    vg.shape.check(locals(), "p2", (3,))

    # Uses "same-side technique" from http://blackpawn.com/texts/pointinpoly/default.html
    along_line = b - a
    return vg.dot(vg.cross(along_line, p1 - a), vg.cross(along_line, p2 - a)) >= 0



import numpy as np
import vg


class Polyline(object):
    """
    Represent the geometry of a polygonal chain in 3-space. The chain may be
    open or closed.

    There are no constraints on the geometry. For example, the chain may be
    simple or self-intersecting, and the points need not be unique.

    The methods do not mutate; they create new polylines which exhibit the
    requested mutation. However, immutability is not enforced. If you wish
    you can mutate a polyline by updating `polyline.v` or `polyline.is_closed`.
    """

    def __init__(self, v, is_closed=False):
        """
        v: np.array containing points in 3-space.
        is_closed: True indicates a closed chain, which has an extra
          segment connecting the last point back to the first
          point.

        """
        # Avoid invoking _update_edges before setting closed and v.
        self.__dict__["is_closed"] = is_closed
        self.v = v

    @classmethod
    def join(cls, *polylines, is_closed=False):
        """
        Join together a stack of open polylines end-to-end into one
        contiguous polyline. The last vertex of the first polyline is
        connected to the first vertex of the second polyline, and so on.
        """
        if len(polylines) == 0:
            raise ValueError("Need at least one polyline to join")
        if any([polyline.is_closed for polyline in polylines]):
            raise ValueError("Expected input polylines to be open, not closed")
        return cls(
            np.vstack([polyline.v for polyline in polylines]), is_closed=is_closed
        )

    def __repr__(self):
        if self.v is not None and self.num_v != 0:
            if self.is_closed:
                return "<closed Polyline with {} verts>".format(self.num_v)
            else:
                return "<open Polyline with {} verts>".format(self.num_v)
        else:
            return "<Polyline with no verts>"

    def __len__(self):
        return self.num_v

    @property
    def num_v(self):
        """
        Return the number of vertices in the polyline.
        """
        return len(self.v)

    @property
    def num_e(self):
        """
        Return the number of segments in the polyline.
        """
        return len(self.e)

    def copy(self):
        """
        Return a copy of this polyline.

        """
        v = None if self.v is None else np.copy(self.v)
        return self.__class__(v, is_closed=self.is_closed)

    def to_dict(self, decimals=3):
        return {
            "vertices": [np.around(v, decimals=decimals).tolist() for v in self.v],
            "edges": self.e,
        }

    def _update_edges(self):
        if self.v is None:
            self.__dict__["e"] = None
            return

        num_vertices = self.v.shape[0]
        num_edges = num_vertices if self.is_closed else num_vertices - 1

        edges = np.vstack([np.arange(num_edges), np.arange(num_edges) + 1]).T

        if self.is_closed:
            edges[-1][1] = 0

        edges.flags.writeable = False

        self.__dict__["e"] = edges

    @property
    def v(self):
        return self.__dict__["v"]

    @v.setter
    def v(self, new_v):
        """
        Update the vertices to a new array-like thing containing points
        in 3D space. Set to None for an empty polyline.

        """
        if new_v is not None:
            vg.shape.check_value(new_v, (-1, 3))
        self.__dict__["v"] = new_v
        self._update_edges()

    @property
    def is_closed(self):
        return self.__dict__["is_closed"]

    @is_closed.setter
    def is_closed(self, new_is_closed):
        """
        Update whether the polyline is closed or open.

        """
        self.__dict__["is_closed"] = new_is_closed
        self._update_edges()

    @property
    def e(self):
        """
        Return the edges of the polyline: an array containing a pair of
        vertex indices for each edge. This is derived automatically from
        `self.v` and `self.is_closed` whenever those values are set.
        """
        return self.__dict__["e"]

    @property
    def segments(self):
        """
        Coordinate pairs for each segment.
        """
        return self.v[self.e]

    @property
    def segment_lengths(self):
        """
        The length of each of the segments.

        """
        if self.e is None:
            return np.zeros(0)
        else:
            v1s = self.v[self.e[:, 0]]
            v2s = self.v[self.e[:, 1]]
            return vg.euclidean_distance(v1s, v2s)

    @property
    def total_length(self):
        """
        The total length of all the segments.

        """
        return np.sum(self.segment_lengths)

    @property
    def segment_vectors(self):
        """
        Vectors spanning each segment.
        """
        segments = self.segments
        return segments[:, 1] - segments[:, 0]

    @property
    def bounding_box(self):
        """
        The bounding box which encompasses the entire polyline.
        """
        from .. import Box

        if self.num_v == 0:
            return None

        return Box.from_points(self.v)

    def index_of_vertex(self, point, atol=1e-08):
        """
        Return the index of the vertex with the given point. If there are
        coincident vertices at that point, return the one at the lowest
        index.
        """
        vg.shape.check(locals(), "point", (3,))

        (matching_indices,) = (
            np.isclose(self.v - point, 0, atol=atol).all(axis=1).nonzero()
        )

        try:
            return matching_indices[0]
        except IndexError:
            # `pass` before `raise` to avoid propagating the IndexError.
            pass
        raise ValueError("No matching vertex")

    def with_insertions(self, points, indices, ret_new_indices=False):
        """
        Return a new polyline with the given points inserted before the given
        indices.

        With `ret_new_indices=True`, also returns the new indices of the
        original vertices and the new indices of the inserted points.
        """
        k = vg.shape.check(locals(), "points", (-1, 3))
        vg.shape.check(locals(), "indices", (k,))

        new_polyline = Polyline(
            v=np.insert(self.v, indices, points, axis=0), is_closed=self.is_closed,
        )

        if not ret_new_indices:
            return new_polyline

        # Compute indices of original vertices.
        old_num_v = self.num_v
        stepwise_index_offsets = np.zeros(old_num_v, dtype=np.int64)
        stepwise_index_offsets[indices[indices < old_num_v]] = 1
        cumulative_index_offsets = np.cumsum(stepwise_index_offsets)
        indices_of_original_vertices = np.arange(old_num_v) + cumulative_index_offsets

        # Compute indices of inserted points. When more than one point is
        # inserted, this will differ from `indices`.
        # TODO: I think this will cause an IndexError when new points are
        # inserted at the end. `indices + cumulative_index_offsets[indices - 1]`
        # would work instead, but would produce an incorrect result for points
        # inserted at the beginning.
        indices_of_inserted_points = indices + cumulative_index_offsets[indices] - 1

        return new_polyline, indices_of_original_vertices, indices_of_inserted_points

    def flipped(self):
        """
        Flip the polyline from end to end. Return a new polyline.
        """
        return Polyline(v=np.flipud(self.v), is_closed=self.is_closed)

    def aligned_with(self, vector):
        """
        Flip the polyline if necessary, so it's aligned with the given
        vector rather than against it. Works on open polylines and considers
        only the two end vertices.
        """
        if self.is_closed:
            raise ValueError("Can't align a closed polyline")

        vg.shape.check(locals(), "vector", (3,))

        if self.num_v < 2:
            return self

        extent = self.v[-1] - self.v[0]
        projected = vg.project(extent, onto=vector)
        if vg.scale_factor(projected, vector) < 0:
            return self.flipped()
        else:
            return self

    def rolled(self, index, ret_edge_mapping=False):
        """
        Return a new Polyline which reindexes the callee polyline, which much
        be closed, so the vertex with the given index becomes vertex 0.

        ret_edge_mapping: if True, return an array that maps from old edge
            indices to new.
        """
        if not self.is_closed:
            raise ValueError("Can't roll an open polyline")

        result = Polyline(v=np.roll(self.v, -index, axis=0), is_closed=True)

        if ret_edge_mapping:
            edge_mapping = np.roll(np.arange(self.num_v), -index)
            return result, edge_mapping
        else:
            return result

    def subdivided_by_length(self, max_length, ret_indices=False):
        """
        Subdivide each line segment longer than `max_length` with
        equal-length segments, such that none of the new segments are longer
        than `max_length`. Returns a new Polyline.

        Args:
            max_length (float): The maximum lenth of a segment.
            ret_indices (bool): When `True`, also returns the indices of the
                original vertices.
        """
        import itertools
        from ..segment import subdivide_segment

        old_num_e = self.num_e
        old_num_v = self.num_v
        num_segments_needed = np.ceil(self.segment_lengths / max_length).astype(
            dtype=np.int64
        )
        (es_to_subdivide,) = (num_segments_needed > 1).nonzero()
        vs_to_insert = [
            subdivide_segment(
                self.v[self.e[old_e_index][0]],
                self.v[self.e[old_e_index][1]],
                np.int(num_segments_needed[old_e_index]),
                endpoint=False,
            )[
                # Exclude the start point, which like the endpoint, is already
                # present.
                1:
            ]
            for old_e_index in es_to_subdivide
        ]

        splits_of_original_vs = np.vsplit(self.v, es_to_subdivide + 1)
        new_polyline = Polyline(
            v=np.concatenate(
                list(
                    itertools.chain(
                        *zip(
                            splits_of_original_vs,
                            vs_to_insert + [np.empty((0, 3), dtype=np.float64)],
                        )
                    )
                )
            ),
            is_closed=self.is_closed,
        )

        if not ret_indices:
            return new_polyline

        # In a degenerate case, `partition_segment()` may return fewer than
        # the requested number of indices. So, recompute the actual number of
        # segments inserted.
        num_segments_inserted = np.zeros(old_num_e, dtype=np.int64)
        num_segments_inserted[es_to_subdivide] = [len(vs) for vs in vs_to_insert]
        stepwise_index_offsets = np.concatenate(
            [
                # The first vertex is never moved.
                np.zeros(1, dtype=np.int64),
                # In a closed polyline, the last edge goes back to vertex
                # 0. Subdivisions of that segment do not affect indexing of
                # any of the vertices (since the original end vertex is
                # still at index 0).
                num_segments_inserted[:-1] if self.is_closed else num_segments_inserted,
            ]
        )
        cumulative_index_offsets = np.sum(
            np.tril(np.broadcast_to(stepwise_index_offsets, (old_num_v, old_num_v))),
            axis=1,
        )
        indices_of_original_vertices = np.arange(old_num_v) + cumulative_index_offsets
        return new_polyline, indices_of_original_vertices

    def with_segments_bisected(self, segment_indices, ret_new_indices=False):
        """
        Return a new polyline with the given segments cut in half.

        With `ret_new_indices=True`, also returns the new indices of the
        original vertices and the new indices of the inserted points.
        """
        return self.with_insertions(
            points=np.mean(self.segments[segment_indices], axis=0),
            indices=self.e[segment_indices][:, 1],
            ret_new_indices=ret_new_indices,
        )

    def apex(self, axis):
        """
        Find the most extreme point in the direction of the axis provided.

        axis: A vector, which is an 3x1 np.array.

        """
        return vg.apex(self.v, axis)

    def intersect_plane(self, plane, ret_edge_indices=False):
        """
        Returns the points of intersection between the plane and any of the
        edges of `polyline`, which should be an instance of Polyline.

        TODO: This doesn't correctly handle vertices which lie on the plane.
        """
        # TODO: Refactor to use `..plane.intersections.intersect_segment_with_plane()`.
        # Identify edges with endpoints that are not on the same side of the plane
        signed_distances = plane.signed_distance(self.v)
        which_es = np.abs(np.sign(signed_distances)[self.e].sum(axis=1)) != 2
        # For the intersecting edges, compute the distance of the endpoints to the plane
        endpoint_distances = np.abs(signed_distances[self.e[which_es]])
        # Normalize the rows of endpoint_distances
        t = endpoint_distances / endpoint_distances.sum(axis=1)[:, np.newaxis]
        # Take a weighted average of the endpoints to obtain the points of intersection
        intersection_points = (
            (1.0 - t[:, :, np.newaxis]) * self.segments[which_es]
        ).sum(axis=1)
        if ret_edge_indices:
            return intersection_points, which_es.nonzero()[0]
        else:
            return intersection_points

    def sliced_by_plane(self, plane):
        """
        Return a new Polyline which keeps only the part that is in front of the given
        plane.

        For open polylines, the plane must intersect the polyline exactly once.

        For closed polylines, the plane must intersect the polyline exactly
        twice, leaving a single contiguous segment in front.
        """
        from ._slice_by_plane import slice_open_polyline_by_plane

        if self.is_closed and self.num_v > 1:
            signed_distances = plane.signed_distance(self.v)
            signs_of_verts = np.sign(signed_distances)
            # For closed polylines, roll the edges so the ones in front of the
            # plane start at index 1 and the one to be cut is at index 0. (If
            # that edge stops directly on the plane, it may not actually need
            # to be cut.) This reduces it to the open polyline intersection
            # problem.
            if signs_of_verts[-1] == 1:
                # e.g. signs_of_verts = np.array([1, -1, -1, 1, 1, 1, 1])
                (vertices_not_in_front,) = np.where(signs_of_verts != 1)
                roll = -vertices_not_in_front[-1]
            else:
                # e.g. signs_of_verts = np.array([-1, 1, 1, 1, 1, 1, -1, -1])
                (vertices_in_front,) = np.where(signs_of_verts == 1)
                if len(vertices_in_front) > 0:
                    roll = -vertices_in_front[0] + 1
                else:
                    roll = 0
            working_v = np.roll(self.v, roll, axis=0)
        else:
            working_v = self.v

        return Polyline(
            v=slice_open_polyline_by_plane(working_v, plane), is_closed=False
        )

    def sliced_at_indices(self, start, stop):
        """
        Take an slice of the given polyline starting at the `start` vertex
        index and ending just befeor reaching the `stop` vertex index. Always
        returns an open polyline.

        When called on a closed polyline, the indies can wrap around the end.
        """
        if stop <= start:
            if self.is_closed:
                num_to_keep = len(self.v) - start + stop
                working_v = np.roll(self.v, -start, axis=0)[0:num_to_keep]
            else:
                raise ValueError(
                    "For an open polyline, start index of slice should be less than stop index"
                )
        else:
            working_v = self.v[start:stop]
        return Polyline(v=working_v, is_closed=False)

    def nearest(self, points, ret_segment_indices=False):
        """
        For the given query point or points, return the nearest point on the
        polyline. With `ret_segment_indices=True`, also return the segment
        indices of those points.
        """
        from .._common.shape import columnize
        from ..segment import closest_point_of_line_segment

        points, _, transform_result = columnize(points, name="points")
        num_points = len(points)

        stacked_points = np.repeat(points, self.num_e, axis=0)
        closest_points_of_segments = closest_point_of_line_segment(
            points=stacked_points,
            start_points=np.tile(self.segments[:, 0], (num_points, 1)),
            segment_vectors=np.tile(self.segment_vectors, (num_points, 1)),
        )
        distance_to_closest_points_of_segments = vg.euclidean_distance(
            stacked_points, closest_points_of_segments
        )

        closest_points_of_segments = closest_points_of_segments.reshape(
            num_points, self.num_e, 3
        )
        distance_to_closest_points_of_segments = distance_to_closest_points_of_segments.reshape(
            num_points, self.num_e
        )

        indices_of_nearest_segments = np.argmin(
            distance_to_closest_points_of_segments, axis=1
        )
        closest_points_of_polyline = np.take_along_axis(
            closest_points_of_segments,
            indices_of_nearest_segments.reshape(num_points, 1, 1),
            axis=1,
        ).reshape(num_points, 3)

        if ret_segment_indices:
            return (
                transform_result(closest_points_of_polyline),
                transform_result(indices_of_nearest_segments),
            )
        else:
            return transform_result(closest_points_of_polyline)

    def sliced_at_points(self, start_point, end_point, atol=1e-8):
        """
        Take a slice of the given polyline at the given start and end points.
        These are expected to be on a vertex or on a segment. If on a segment
        (or near to but not directly on a segment) a new point is inserted
        at exactly the given point.
        """
        vg.shape.check(locals(), "start_point", (3,))
        vg.shape.check(locals(), "end_point", (3,))

        working_polyline = self

        try:
            # Check if the start point intersects a vertex. If it does, great;
            # if not, insert it.
            start_v_index = working_polyline.index_of_vertex(start_point)
        except ValueError:
            nearest_point, segment_index = working_polyline.nearest(
                start_point, ret_segment_indices=True
            )
            (_, start_v_index) = working_polyline.e[segment_index]
            working_polyline = working_polyline.with_insertions(
                points=nearest_point.reshape(-1, 3), indices=np.array([start_v_index]),
            )

        try:
            end_v_index = working_polyline.index_of_vertex(end_point)
        except ValueError:
            nearest_point, segment_index = working_polyline.nearest(
                end_point, ret_segment_indices=True
            )
            (_, end_v_index) = working_polyline.e[segment_index]
            (
                working_polyline,
                indices_of_original_vertices,
                _,
            ) = working_polyline.with_insertions(
                points=nearest_point.reshape(-1, 3),
                indices=np.array([end_v_index]),
                ret_new_indices=True,
            )
            start_v_index = indices_of_original_vertices[start_v_index]

        # Then slice at those points.
        return working_polyline.sliced_at_indices(start_v_index, end_v_index + 1)



from . import _plane_functions
from ._plane_functions import *  # noqa: F401,F403
from ._plane_intersect import intersect_segment_with_plane

__all__ = _plane_functions.__all__ + ["intersect_segment_with_plane"]



import numpy as np
import vg


def inflection_points(points, axis, span):
    """
    Find the list of vertices that preceed inflection points in a curve. The curve is differentiated
    with respect to the coordinate system defined by axis and span.

    Interestingly, `lambda x: 2*x + 1` should have no inflection points, but
    almost every point on the line is detected. It's because a zero or zero
    crossing in the second derivative is necessary but not sufficient to
    detect an inflection point. You also need a higher derivative of odd
    order that's non-zero. But that gets ugly to detect reliably using sparse
    finite differences. Just know that if you've got a straight line this
    method will go a bit haywire.

    axis: A vector representing the vertical axis of the coordinate system.
    span: A vector representing the the horiztonal axis of the coordinate system.

    returns: a list of points in space corresponding to the vertices that
    immediately preceed inflection points in the curve
    """
    vg.shape.check(locals(), "points", (-1, 3))
    vg.shape.check(locals(), "axis", (3,))
    vg.shape.check(locals(), "span", (3,))

    coords_on_span = points.dot(span)
    coords_on_axis = points.dot(axis)

    # Take the second order finite difference of the curve with respect to the
    # defined coordinate system
    finite_difference_1 = np.gradient(coords_on_axis, coords_on_span)
    finite_difference_2 = np.gradient(finite_difference_1, coords_on_span)

    # Compare the product of all neighboring pairs of points in the second
    # derivative. If a pair of points has a negative product, then the second
    # derivative changes sign between those points. These are the inflection
    # points.
    is_inflection_point = np.concatenate(
        [finite_difference_2[:-1] * finite_difference_2[1:] <= 0, [False]]
    )

    return points[is_inflection_point]



import numbers
import numpy as np
import vg
from .._common.shape import check_shape_any, columnize

__all__ = [
    "plane_normal_from_points",
    "plane_equation_from_points",
    "normal_and_offset_from_plane_equations",
    "signed_distance_to_plane",
    "project_point_to_plane",
]


def plane_normal_from_points(points, normalize=True):
    """
    Given a set of three points, compute the normal of the plane which
    passes through them. Also works on stacked inputs (i.e. many sets
    of three points).

    This is the same as `polliwog.tri.functions.surface_normals`, to
    which this delegates.
    """
    from ..tri import surface_normals

    return surface_normals(points=points, normalize=normalize)


def plane_equation_from_points(points):
    """
    Given many sets of three points, return a stack of plane equations
    [`A`, `B`, `C`, `D`] which satisfy `Ax + By + Cz + D = 0`. Also
    works on three points to return a single plane equation.

    These coefficients can be decomposed into the plane normal vector
    which is `[A, B, C]` and the offset `D`, either by the caller or
    by using `normal_and_offset_from_plane_equations()`.
    """
    points, _, transform_result = columnize(points, (-1, 3, 3), name="points")

    p1s = points[:, 0]
    unit_normals = plane_normal_from_points(points)
    D = -vg.dot(p1s, unit_normals)

    return transform_result(np.hstack([unit_normals, D.reshape(-1, 1)]))


def normal_and_offset_from_plane_equations(plane_equations):
    """
    Given `A`, `B`, `C`, `D` of the plane equation `Ax + By + Cz + D = 0`,
    return the plane normal vector which is `[A, B, C]` and the offset `D`.
    """
    check_shape_any(plane_equations, (4,), (-1, 4), name="plane_equations")

    if plane_equations.ndim == 2:
        normal = plane_equations[:, :3]
        offset = plane_equations[:, 3]
    else:
        normal = plane_equations[:3]
        offset = plane_equations[3]
    return normal, offset


def signed_distance_to_plane(points, plane_equations):
    """
    Return the signed distances from each point to the corresponding plane.

    For convenience, can also be called with a single point and a single
    plane.
    """
    k = check_shape_any(points, (3,), (-1, 3), name="points")
    check_shape_any(
        plane_equations, (4,), (-1 if k is None else k, 4), name="plane_equations"
    )

    normals, offsets = normal_and_offset_from_plane_equations(plane_equations)
    return vg.dot(points, normals) + offsets


def translate_points_along_plane_normal(points, plane_equations, factor):
    """
    Translate each point along the plane normal. The distance translated
    is specified in terms of its initial distance from the plane.

    Args:
        factor (float): Transform this multiple of the normal. With `0`,
            this returns the identity. With `1`, this projects to the
            plane. When `2` it returns the opposite point.
    """
    k = check_shape_any(points, (3,), (-1, 3), name="points")
    check_shape_any(
        plane_equations, (4,), (-1 if k is None else k, 4), name="plane_equations"
    )
    assert isinstance(factor, numbers.Real)

    # Translate the point back to the plane along the normal.
    signed_distance = signed_distance_to_plane(points, plane_equations)
    normals, _ = normal_and_offset_from_plane_equations(plane_equations)

    if np.isscalar(signed_distance):
        return points - factor * signed_distance * normals
    else:
        return points - factor * signed_distance.reshape(-1, 1) * normals


def project_point_to_plane(points, plane_equations):
    """
    Project each point to the corresponding plane.
    """
    k = check_shape_any(points, (3,), (-1, 3), name="points")
    check_shape_any(
        plane_equations, (4,), (-1 if k is None else k, 4), name="plane_equations"
    )

    return translate_points_along_plane_normal(
        points=points, plane_equations=plane_equations, factor=1
    )


def mirror_point_across_plane(points, plane_equations):
    """
    Mirror each point to the corresponding point on the opposite side of the
    plane.
    """
    k = check_shape_any(points, (3,), (-1, 3), name="points")
    check_shape_any(
        plane_equations, (4,), (-1 if k is None else k, 4), name="plane_equations"
    )

    return translate_points_along_plane_normal(
        points=points, plane_equations=plane_equations, factor=2
    )



from . import _line_functions as _functions
from ._line_functions import *  # noqa: F401,F403
from ._line_intersect import intersect_2d_lines, intersect_lines  # noqa: F401

__all__ = ["intersect_lines", "intersect_2d_lines"] + _functions.__all__



import numpy as np


def intersect_lines(p0, q0, p1, q1):
    """
    Intersect two lines in 3d: (p0, q0) and (p1, q1). Each should be a 3D
    point.
    See this for a diagram: http://math.stackexchange.com/questions/270767/find-intersection-of-two-3d-lines
    """
    e = p0 - q0  # direction of line 0
    f = p1 - q1  # direction of line 1

    # Check for special case where we're given the intersection
    # Note that we must check for these, because if p0 == p1 then
    # g would be zero length and we can't continue
    if np.all(p0 == p1) or np.all(p0 == q1):
        return p0
    if np.all(q0 == p1) or np.all(p0 == q1):
        return q0

    g = p0 - p1  # line between to complete a triangle
    h = np.cross(f, g)
    k = np.cross(f, e)
    h_ = np.linalg.norm(h)
    k_ = np.linalg.norm(k)
    if h_ == 0 or k_ == 0:
        # There is no intesection; either parallel (k=0) or collinear (both=0) lines.
        return None

    l = h_ / k_ * e  # noqa: E741 (FIXME)
    sign = -1 if np.all(h / h_ == k / k_) else +1
    return p0 + sign * l


def intersect_2d_lines(p0, q0, p1, q1):
    """
    Intersect two lines: (p0, q0) and (p1, q1). Each should be a 2D
    point.
    """
    # Adapted from http://stackoverflow.com/a/26416320/893113
    dy = q0[1] - p0[1]
    dx = q0[0] - p0[0]
    lhs0 = [-dy, dx]
    rhs0 = p0[1] * dx - dy * p0[0]

    dy = q1[1] - p1[1]
    dx = q1[0] - p1[0]
    lhs1 = [-dy, dx]
    rhs1 = p1[1] * dx - dy * p1[0]

    a = np.array([lhs0, lhs1])
    b = np.array([rhs0, rhs1])

    try:
        return np.linalg.solve(a, b)
    except np.linalg.LinAlgError:
        return None



import vg


class Line:
    def __init__(self, point, along, assume_normalized=False):
        vg.shape.check(locals(), "point", (3,))
        vg.shape.check(locals(), "along", (3,))

        if vg.almost_zero(along):
            raise ValueError("along should not be the zero vector")

        self.reference_point = point
        self.along = along
        self.assume_normalized = assume_normalized

    @classmethod
    def from_points(cls, p1, p2):
        vg.shape.check(locals(), "p1", (3,))
        vg.shape.check(locals(), "p2", (3,))
        return cls(point=p1, along=p2 - p1)

    @property
    def reference_points(self):
        """
        Return two reference points on the line.
        """
        return self.reference_point, self.reference_point + self.along

    def intersect_line(self, other):
        """
        Find the intersection with another line.
        """
        from ._line_intersect import intersect_lines

        return intersect_lines(*(self.reference_points + other.reference_points))

    def project(self, points):
        """
        Project a given point (or stack of points) to the plane.
        """
        from ._line_functions import project_point_to_line

        return project_point_to_line(
            points=points,
            reference_points_of_lines=self.reference_point,
            vectors_along_lines=self.along,
        )



import numpy as np
import vg


def slice_open_polyline_by_plane(vertices, plane):
    from .. import Plane
    from ..plane import intersect_segment_with_plane

    num_v = vg.shape.check(locals(), "vertices", (-1, 3))
    if num_v == 0:
        raise ValueError("A plane can't intersect a polyline with no points")
    if not isinstance(plane, Plane):
        raise ValueError("plane should be an instance of polliwog.Plane")

    signed_distances = plane.signed_distance(vertices)
    signs_of_vertices = np.sign(signed_distances)

    (transition_points,) = (signs_of_vertices[:-1] != signs_of_vertices[1:]).nonzero()
    components = np.vsplit(vertices, transition_points + 1)
    component_signs = signs_of_vertices[np.concatenate([[0], transition_points + 1])]

    (components_in_front,) = (component_signs == 1).nonzero()
    if len(components_in_front) == 0:
        raise ValueError("Polyline has no vertices in front of the plane")
    elif len(components_in_front) > 1:
        raise ValueError("Polyline intersects the plane too many times")
    elif len(components) < 2:
        raise ValueError("Polyline lies entirely in front of the plane")
    (component_in_front,) = components_in_front
    verts_in_front = components[component_in_front]

    # Prepend the plane intersection point with the previous component.
    if component_in_front > 0:
        adjacent_vertex = components[component_in_front - 1][-1]
        sign_of_adjacent_vertex = component_signs[component_in_front - 1]
        if sign_of_adjacent_vertex == 0:
            prepend = adjacent_vertex
        else:
            prepend = intersect_segment_with_plane(
                start_points=adjacent_vertex,
                segment_vectors=verts_in_front[0] - adjacent_vertex,
                points_on_plane=plane.reference_point,
                plane_normals=plane.normal,
            )
    else:
        prepend = np.zeros((0, 3))

    # Append the plane intersection point with the next component.
    if component_in_front + 1 < len(components):
        adjacent_vertex = components[component_in_front + 1][0]
        sign_of_adjacent_vertex = component_signs[component_in_front + 1]
        if sign_of_adjacent_vertex == 0:
            append = adjacent_vertex
        else:
            last_vert = verts_in_front[-1]
            append = intersect_segment_with_plane(
                start_points=last_vert,
                segment_vectors=adjacent_vertex - last_vert,
                points_on_plane=plane.reference_point,
                plane_normals=plane.normal,
            )
    else:
        append = np.zeros((0, 3))

    return np.vstack([prepend, verts_in_front, append])



"""
Functions for creating sets of triangles to model 3D shapes.

These functions have two possible return types:

- When `ret_unique_vertices_and_faces=True`, they return a vertex array (with
  each vertex listed once) and a face array (i.e. an array of triples of vertex
  indices). This is ideal when using with a mesh library like Lace
  (https://github.com/lace/lace/) or Trimesh (https://trimsh.org/) or when you
  care about the topology.
- When `ret_unique_vertices_and_faces=False`, they return a flattened array
  of triangle coordinates with each vertex repeated. This is useful for
  computation that use flattened triangle coordinates, such as the functions
  in `polliwog.tri`.

See also:
    https://en.wikipedia.org/wiki/Tessellation_(computer_graphics)
"""

from . import _shapes
from ._shapes import *  # noqa: F401,F403

__all__ = _shapes.__all__



import numpy as np
import vg
from .._common.shape import columnize
from ..line._line_functions import coplanar_points_are_on_same_side_of_line

__all__ = [
    "surface_normals",
    "tri_contains_coplanar_point",
    "barycentric_coordinates_of_points",
]


def surface_normals(points, normalize=True):
    """
    Compute the surface normal of a triangle. The direction of the normal
    follows conventional counter-clockwise winding and the right-hand
    rule.

    Also works on stacked inputs (i.e. many sets of three points).
    """
    points, _, transform_result = columnize(points, (-1, 3, 3), name="points")

    p1s = points[:, 0]
    p2s = points[:, 1]
    p3s = points[:, 2]
    v1s = p2s - p1s
    v2s = p3s - p1s
    normals = vg.cross(v1s, v2s)

    if normalize:
        normals = vg.normalize(normals)

    return transform_result(normals)


def tri_contains_coplanar_point(a, b, c, point):
    """
    Assuming `point` is coplanar with the triangle `ABC`, check if it lies
    inside it.
    """
    vg.shape.check(locals(), "a", (3,))
    vg.shape.check(locals(), "b", (3,))
    vg.shape.check(locals(), "c", (3,))
    vg.shape.check(locals(), "point", (3,))

    # Uses "same-side technique" from http://blackpawn.com/texts/pointinpoly/default.html
    return (
        coplanar_points_are_on_same_side_of_line(b, c, point, a)
        and coplanar_points_are_on_same_side_of_line(a, c, point, b)
        and coplanar_points_are_on_same_side_of_line(a, b, point, c)
    )


def barycentric_coordinates_of_points(vertices_of_tris, points):
    """
    Compute barycentric coordinates for the projection of a set of points to a
    given set of triangles specfied by their vertices.

    These barycentric coordinates can refer to points outside the triangle.
    This happens when one of the coordinates is negative. However they can't
    specify points outside the triangle's plane. (That requires tetrahedral
    coordinates.)

    The returned coordinates supply a linear combination which, applied to the
    vertices, returns the projection of the original point the plane of the
    triangle.

    Args:
        vertices_of_tris (np.arraylike): A set of triangle vertices as `kx3x3`.
        points (np.arraylike): Coordinates of points as `kx3`.

    Returns:
        np.ndarray: Barycentric coordinates as `kx3`

    See Also:
        - https://en.wikipedia.org/wiki/Barycentric_coordinate_system
        - Heidrich, "Computing the Barycentric Coordinates of a Projected
          Point," JGT 05 (http://www.cs.ubc.ca/~heidrich/Papers/JGT.05.pdf)
    """
    k = vg.shape.check(locals(), "vertices_of_tris", (-1, 3, 3))
    vg.shape.check(locals(), "points", (k, 3))

    p = points.T
    q = vertices_of_tris[:, 0].T
    u = (vertices_of_tris[:, 1] - vertices_of_tris[:, 0]).T
    v = (vertices_of_tris[:, 2] - vertices_of_tris[:, 0]).T

    n = np.cross(u, v, axis=0)
    s = np.sum(n * n, axis=0)

    # If the triangle edges are collinear, cross-product is zero,
    # which makes "s" 0, which gives us divide by zero. So we
    # make the arbitrary choice to set s to epsv (=numpy.spacing(1)),
    # the closest thing to zero
    s[s == 0] = np.spacing(1)

    oneOver4ASquared = 1.0 / s
    w = p - q
    b2 = np.sum(np.cross(u, w, axis=0) * n, axis=0) * oneOver4ASquared
    b1 = np.sum(np.cross(w, v, axis=0) * n, axis=0) * oneOver4ASquared
    b = np.vstack((1 - b1 - b2, b1, b2))

    return b.T
